use crate::{db, DistroMgr};

use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    sync::Arc,
};

use sea_orm::{DbConn, EntityTrait};
use tokio::sync::Mutex;

#[derive(Clone)]
pub struct MetaDistroMgr {
    distro_dir: PathBuf,
    conn: DbConn,
    distros: Arc<Mutex<HashMap<String, Arc<DistroMgr>>>>,
}

impl MetaDistroMgr {
    pub async fn new<P: AsRef<Path>>(distro_dir: P, conn: DbConn) -> crate::Result<Self> {
        if !tokio::fs::try_exists(&distro_dir).await? {
            tokio::fs::create_dir(&distro_dir).await?;
        }

        let distro_dir = distro_dir.as_ref().to_path_buf();
        let mut map: HashMap<String, Arc<DistroMgr>> = HashMap::new();

        let distros = db::Distro::find().all(&conn).await?;

        for distro in distros {
            let mgr =
                DistroMgr::new(distro_dir.join(&distro.name), distro.id, conn.clone()).await?;
            map.insert(distro.name, Arc::new(mgr));
        }

        Ok(Self {
            distro_dir,
            conn,
            distros: Arc::new(Mutex::new(map)),
        })
    }

    pub async fn get_mgr(&self, distro: &str) -> Option<Arc<DistroMgr>> {
        let map = self.distros.lock().await;

        map.get(distro).map(|mgr| Arc::clone(mgr))
    }

    pub async fn get_or_create_mgr(&self, distro: &str) -> crate::Result<Arc<DistroMgr>> {
        let mut map = self.distros.lock().await;

        if let Some(mgr) = map.get(distro) {
            Ok(Arc::clone(mgr))
        } else {
            let distro = db::query::distro::insert(&self.conn, distro, None).await?;

            let mgr = Arc::new(
                DistroMgr::new(
                    self.distro_dir.join(&distro.name),
                    distro.id,
                    self.conn.clone(),
                )
                .await?,
            );
            map.insert(distro.name, Arc::clone(&mgr));

            Ok(mgr)
        }
    }
}
